#!/usr/bin/perl

use strict;
use warnings;

use threads;
use threads::shared;


#use locale;
use Getopt::Long;
use WWW::Telegram::BotAPI;
use File::Basename;
use File::Copy;
use LWP::Simple;
use File::Spec;
# qw 'rel2abs catfile';
use URL::Encode 'url_encode';
use Encode qw 'from_to encode decode';
use String::CRC32;
use uue;
use MIME::Base64;
use File::Flock::Tiny;
use Time::TZOffset qw/tzoffset/;
#use File::Path qw(make_path remove_tree);

our $vers = '0.7.1';

our ( $sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst, $cfgh,$line,
    $api_url, $bot_token, $lockfile, $pktdir, $channelsconf, $logfile,
    $address, $basedirname, $slash, %tags_id, %areaname, $VirtualAddress,
    $uplink, $pktpassword, %options, $temp_dir, $error_message, $fname,
    $result, $fulfilename, $caption, $api, $fnu, $urlonly, $fileonly,
    $filesize, $tg_msg_repl_to, $pkthdr, $oldcr, $badname, $version, $pktver,
    $hdr, $dateTime, $attr,$cost, $ToName, $FromName, $Subj, $Text, $fromchrs,
    $area, $origAddr, $destAddr, $msgid_addr, $msgid_crc, $pos, $len, $domain,
    $ldomain, $userconf, %realnames, $updcount, $u, $updateid, $chatid,
    $msgtext, $nu, $forward_from_chat_title, $tg_msg_id, $reply_msgid,
    $chattitle, $replytext, $fromname, $fwd_fromname, $from_user_id, $toname,
    $toaddr, $subj, $to_user_id, $quotechars, $outpktdir, $get_file_url,
    $file_buffer, $local_url, $web_files_path, %usernames, $origin,
    $notifications, $fidomsgid, $msg_crc, $LFH, $cycle_time, $cycle_count,
    $dupe_file, $dupe_dir, $DFH, %dupe_msgids, $echo_tag, $helpfile,
    $helptext, $az,$an,$af,$ap, %UserProfile, $UsersProfileDir, $exitflag,
    $conf_size, $conf_mtime );
    
our $tearline = "--- tg BBS v$vers\r";

our $TZUTC = sprintf( "%04d", tzoffset( localtime() ) );
#   $TZUTC = s/\+//;

$0 =~ /[\\\/]+([^\\\/]+)$/;
my $usage = "Usage: $1 --config=/path/filename.conf\n".
		"Options: --toss\n".
		"         --scan\n";

our ( $config_file, $toss, $scan );

GetOptions ("config=s" => \$config_file,
            "toss"     => \$toss,   # flag
            "scan"     => \$scan   # flag
	    )
or die("Error in command line arguments\n");

if ( !defined( $config_file ) ) {
	print STDERR "\nConfig file MUST be defined.\n\n$usage";
	exit;
}

my ($channels_size, $channels_time);

sub conf_stat()
{
    my $c_s = (stat( $config_file ))[7];
    my $c_t = (stat( $config_file ))[9];
    $conf_size = $c_s unless defined $conf_size;
    $conf_mtime = $c_t unless defined $conf_mtime;
    if ( $conf_size != $c_s || $conf_mtime != $c_t ) {
#	read_conf();
	return 1;
    }
    $c_s = (stat( $channelsconf ))[7];
    $c_t = (stat( $channelsconf ))[9];
    if ( $channels_size != $c_s || $channels_time != $c_t ) {
#	read_conf();
	return 1;
    }

return 0;
}

sub read_conf()
{
	print "Reading config. $config_file\n";
	writelog("Reading config. $config_file");
	File::Spec->rel2abs( __FILE__ ) =~ /(.*)([\\\/])[^\\\/]+$/;
	$basedirname = $1;
	$slash = $2;
	if( open( $cfgh, "<$config_file" ) ){
		while( $line = <$cfgh> ){
			$line =~ s/\r//g;
			$line =~ s/\n//g;
			$line =~ s/	/ /g while $line =~ /	/;
			$line =~ s/^ //g while $line =~ /^ /;
			next if $line =~ /^\#.*/;

			if ( $line =~ /^bot_token[ ]+([^ \r]+)/i ) {
				$bot_token = $1;
			} elsif( $line =~ /^api_url[ ]+([^ \r]+)/i ) {
				$api_url =$1;
			} elsif( $line =~ /^get_file_url[ ]+([^ \r]+)/i ) {
				$get_file_url =$1;
			} elsif( $line =~ /^web_files_path[ ]+([^ \r]+)/i ) {
				$web_files_path =$1;
			} elsif( $line =~ /^local_url[ ]+([^ \r]+)/i ) {
				$local_url =$1;
			} elsif( $line =~ /^ADDRESS[ ]+([^ \r]+)/i ) {
				$address =$1;
			} elsif( $line =~ /^lockfile[ ]+([^ \r\n]+)/i ) {
				$lockfile =$1;
			} elsif( $line =~ /^Origin[ ]+([^\n\r]+)/i ) {
				$origin =$1;
			} elsif( $line =~ /^in_pktdir[ ]+([^ \r\n]+)$/i ) {
				$pktdir = $1;
				if ( $pktdir =~ /(.*)[\\\/]$/) {
					$pktdir = $1;
				}
			} elsif( $line =~ /^dupe_dir[ ]+([^ \r\n]+)$/i ) {
				$dupe_dir = $1;
				if ( $dupe_dir =~ /(.*)[\\\/]$/) {
					$dupe_dir = $1;
				}
			} elsif( $line =~ /^out_pktdir[ ]+([^ \r\n]+)$/i ) {
				$outpktdir = $1;
				if ( $outpktdir =~ /(.*)[\\\/]$/) {
					$outpktdir = $1;
				}
			} elsif( $line =~ /^channelsconf ([^ \r]+)/i ) {
				$channelsconf =$1;
			} elsif( $line =~ /^logfile[ ]+([^ \r]+)/i ) {
				$logfile =$1;
			} elsif( $line =~ /^uplink[ ]+([^ \r]+)/i ) {
				$uplink = $1;
			} elsif( $line =~ /^VirtualAddress[ ]+([^ \r]+)/i ) {
				$VirtualAddress = $1;
			} elsif( $line =~ /^pktpassword[ ]+([^ \r]+)/i ) {
				$pktpassword = $1;
			} elsif( $line =~ /^temp_dir[ ]+([^ \r]+)/i ) {
				$temp_dir = $1;
			} elsif( $line =~ /^userconf[ ]+([^ \r]+)/i ) {
				$userconf = $1;
			} elsif( $line =~ /^cycle_time\s+(\d+)/i ){
			    $cycle_time = $1;
			} elsif( $line =~ /^cycle_count\s+(\d+)/i ){
			    $cycle_count = $1;
			} elsif( $line =~ /^helpfile\s+([^\s\r\n]+)/i ){
			    $helpfile = $1;
			} elsif( $line =~ /^exit_flag\s+([^\s\r\n]+)/i ){
			    $exitflag = $1;
			} elsif( $line =~ /^UsersProfileDir[ ]+([^ \r\n]+)$/i ){
				$UsersProfileDir = $1;
				if ( $UsersProfileDir =~ /(.*)[\\\/]$/) {
					$UsersProfileDir = $1;
				}
			}

		}
	    close($cfgh);
	    $conf_size = (stat( $config_file ))[7];
	    $conf_mtime = (stat( $config_file ))[9];
	} else {
		print STDERR "Can't open \'$config_file\'. ($!).\n";
		exit;
	}
print "Done.\n";
writelog("Reading config done. \'$config_file\'");
}

sub check_errors()
{
# <--- Error checking --->
		print "Checking config errors...\n";
		if( !defined( $bot_token ) ) {
			print STDERR "ERROR: Bot token MUST be defined in config file.\n";
			exit;
		}
		if( !defined( $api_url ) ) {
			$api_url = 'https://api.telegram.org/bot%s/%s';
			print STDERR "WARNING: API URL not defined. Using default \'https:\/\/api.telegram.org\/bot\%s\/\%s\'.\n";
		}
		if( !defined( $local_url ) ) {
#			$local_url = 'https://api.telegram.org/bot%s/%s';
			print STDERR "WARNING: Local URL is not defined. Files will be sent as UUE.\n";
#			exit;
		} elsif( $local_url !~ /[\\\/]$/ ) {
			$local_url .= '/';
		}
		if( !defined( $get_file_url ) ) {
			$get_file_url = 'https://api.telegram.org/file/bot';
			print STDERR "WARNING: Get File URL not defined. Using default \'https:\/\/api.telegram.org\/file\/bot\'.\n";
		}
		if( !defined( $address ) ) {
			print STDERR "ERROR: System Address MUST be defined in config file.\n";
			exit;
		} elsif ($address =~ /^(\d+)\:(\d+)\/(\d+)\.?(\d*)$/) {
		    ( $az, $an, $af, $ap ) = ( $1, $2, $3, $4 );
		    $ap = 0 unless defined $ap;
		} else {
		    print STDERR "ERROR: System Address MUST be the walid Fidonet addres.\n";
		    exit;
		}
		if( !defined( $VirtualAddress ) ) {
			print STDERR "ERROR: PKT From Address MUST be defined in config file.\n";
			exit;
		} elsif ($VirtualAddress !~ /^\d+\:\d+\/\d+\.?\d*$/) {
			print STDERR "ERROR: PKT From Address MUST be the walid Fidonet addres.\n";
			exit;
		} elsif ($VirtualAddress =~ /^\d+\:\d+\/\d+$/) {
			$VirtualAddress .= '.0';
		}
		if( !defined( $uplink ) ) {
			print STDERR "ERROR: PKT To Address MUST be defined in config file.\n";
			exit;
		} elsif ($uplink !~ /^\d+\:\d+\/\d+\.?\d*$/) {
			print STDERR "ERROR: PKT To Address MUST be the walid Fidonet addres.\n";
			exit;
		} elsif ($uplink =~ /^\d+\:\d+\/\d+$/) {
			$uplink .= '.0';
		}
		if( !defined( $origin ) ){
		    print STDERR "WARNING: Origing not defined. Using default.\n";
		    writelog("WARNING: Origing not defined. Using default.");
		    $origin = "Fido by Telegram BBS by Stas Mishchenkov";
		}
		if( !defined( $lockfile ) ) {
			$lockfile = $basedirname . $slash . 'tg.lock';
			print STDERR "WARNING: Lock file not defined. Using default \'$lockfile\'.\n";
		}
		if( !defined( $pktpassword ) ) {
			print STDERR "WARNING: PKT password not defined. PKT password cheking disabled.\n";
		} elsif ( length( $pktpassword ) != 8 ) {
			print STDERR "ERROR: PKT password MUST be 8 characters long.\n";
			exit;
		}
		if( !defined( $pktdir ) ) {
			print STDERR "ERROR: PKT dir MUST be defined in config file.\n";
			exit;
		} else {
			if ($pktdir =~ /(.*)[\\\/]$/) {
				$pktdir = $1;
			}
			if( !-d $pktdir ){
				print STDERR "ERROR: PKT dir MUST exist.\n";
				exit;
			} 
		}
		if( !defined( $dupe_dir ) ) {
			print STDERR "ERROR: DUPE dir MUST be defined in config file.\n";
			exit;
		} else {
			if ($dupe_dir =~ /(.*)[\\\/]$/) {
				$dupe_dir = $1;
			}
			if( !-d $dupe_dir ){
				print STDERR "ERROR: DUPE dir MUST exist.\n";
				exit;
			}
		}
		if( !defined( $outpktdir ) ) {
			print STDERR "ERROR: Out PKT dir MUST be defined in config file.\n";
			exit;
		} else {
			if ($outpktdir =~ /(.*)[\\\/]$/) {
				$outpktdir = $1;
			}
			if( !-d $outpktdir ){
				print STDERR "ERROR: Out PKT dir MUST exist.\n";
				exit;
			}
		}
		if( !defined( $temp_dir ) ) {
			$temp_dir = $ENV{TMP};
			print STDERR "WARNING: Temp Dir not defined. $ENV{TMP} used instead.\n";
			writelog("WARNING: Temp Dir not defined. $ENV{TMP} used instead.");
		} else {
			if ($temp_dir =~ /(.*)[\\\/]$/) {
				$temp_dir = $1;
			}
			if( !-d $temp_dir ){
				print STDERR "ERROR: Temp dir MUST exist.\n";
				exit;
			}
		}
		if( !defined($web_files_path) ) {
		    $web_files_path = $temp_dir;
		    print STDERR "WARNING: web_files_path is not defined. $temp_dir used instead.\n";
		    writelog("WARNING: web_files_path is not defined. $temp_dir used instead.");
		} else {
		    $web_files_path = $1 if $web_files_path =~ /(.*)[\\\/]$/;
		}
		if( !defined( $channelsconf ) ) {
			$channelsconf = $basedirname . $slash . 'channels.conf';
			print STDERR "WARNING: Channels conf not defined. Using default \'$channelsconf\'.\n";
		}
		if( !defined( $userconf ) ) {
			$userconf = $basedirname . $slash . 'userconf.conf';
			print STDERR "WARNING: User conf not defined. Using default \'$userconf\'.\n";
		}
		if( !defined( $logfile ) ) {
			print STDERR "WARNING: Log file not defined. No logging will be.\n";
		}
		if( !defined( $cycle_time ) ) {
			$cycle_time = 6;
		}
		if( !defined( $cycle_count ) ) {
		    $cycle_count = 1;
		}
		unless( defined( $helpfile ) || -e $helpfile ) {
		    $helptext = 'Sorry. No help message defined.';
		} else {
		    my $helpsize = -s $helpfile;
#		    print STDERR "Help size: $helpsize\nHelp file: \'$helpfile\'\n";
		    if( open( HFH, "<$helpfile" ) ) {
			read( HFH, $helptext, $helpsize );
			close( HFH );
			$helptext = decode( 'cp866', $helptext );
#			print STDERR "\'$helptext\'\n";
		    } else {
			print STDERR "Can't open \'$helpfile\' ($!).\n";
			$helptext = 'Sorry. No help message defined.';
		    }
		}
		if ( defined( $UsersProfileDir ) ) {
		    unless( -d $UsersProfileDir ||
		    -r $UsersProfileDir || -w $UsersProfileDir  ){
			print STDERR "Warning: Users profile dir is not exist.\n";
			undef $UsersProfileDir;
		    }
		}
#		$UsersProfileDir

#$helpfile
#$helptext
# <--- Error checking --->
print "Done.\n";
}

sub writelog
{
    my ( $str ) = @_ ;

    ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime();
    
    if ( defined($logfile) )
    {
        if ( open( FLOG, ">>$logfile") )
        {
	    print( FLOG ($year+1900)."-". sprintf("%02d",$mon+1)."-".
		sprintf("%02d",$mday)." ". sprintf("%02d",$hour).":".
		sprintf("%02d",$min).":". sprintf("%02d",$sec) . " [$$] $str\n" );
	    close( FLOG );
        }
        else
        {
            printf( STDERR "Can't open $logfile. ($!)\n" );
        }
    }
}

sub readchannelsconf()
{
    print "Reading channels conf. $channelsconf\n";
    my $a;
    if( open( CCF, "<$channelsconf" ) ) {
	while( $line = <CCF> ){
		next if $line =~ /^[ 	]*\#.*/;
		$line =~ s/	/ /g while $line =~ /	/;
		$line =~ /^([^ ]+) ([^ \r\n]+)[ ]?([^\r\n]*)$/;
		$tags_id{$1} = $2;
		$areaname{$2} = $1;
		$options{$1} = $3 if defined $3;
		$a = $1;
		if ( $options{$a} =~ /urlonly/i && $options{$a} =~ /fileonly/i ) {
			close(CCF);
			print STDERR "ERROR: Area: $a\n~~~~~~ Mutually exclusive options: urlonly fileonly\n";
			exit;
		}
	}
	close(CCF);
    } else {
	print STDERR "Can't open $channelsconf. ($!).\n";
	writelog("Can't open $channelsconf. ($!).");
	exit;
#	exit_procedure();
    }
    $channels_size = (stat( $channelsconf ))[7];
    $channels_time = (stat( $channelsconf ))[9];
print "Done.\n";
}

sub readuserconf()
{
	writelog("Reading user config. $userconf");
	print "Reading user config. $userconf\n";

	if (open(CU,"<$userconf" )) {
	    while (defined( $line = <CU>)) {
		next if $line =~ /^[ 	]*\#/;
	        next unless $line =~ /^(.*?) -- (.*?) -- ([^\r\n?]+)/;
		$realnames{$3} = $2 if $2 ne '0';
		$usernames{$3} = $1;
	    }
	    close(CU);
	    writelog("Done.");
	    print "Done.\n";
	} else {
	    print STDERR "Can't open $userconf. ($!).\n";
	    writelog("Can't open $userconf. ($!).\n");
	    exit;
	}
}


sub readpktdir()
{
    my $pktlist = '';
    my @pkts;
    print "Reading PKT dir.\n";
    if ( opendir( PKTD, $pktdir ) ){
	while( readdir( PKTD ) ){
	    if ( $_ =~ /\.pkt/i ){
		print " $_\n";
		$pktlist .= " $_";
	    }
	}
	close( PKTD );
    } else {
	print STDERR "Can't open $pktdir. ($!).\n";
	writelog("Can't open $pktdir. ($!).");
#	exit_procedure();
    }

    @pkts = split( ' ', $pktlist );
    foreach my $pkt ( sort @pkts ) {
	writelog("Reading $pkt");
	readPKT( $pktdir . $slash . $pkt );
    }
print "Reading PKT dir done.\n";
}

sub send_decoded($)
{
	my ( $Chat_ID ) = @_;
	undef $error_message;
	if ( defined($options{$Chat_ID}) && $options{$Chat_ID} =~ /\-dn/i ) {
	    $notifications = 1;
	} else {
	    $notifications = 0;
	}
#$notifications
	if ( $fname =~ /\.(jpg|png)$/i ) {
		writelog("Sending as Photo: $fname");
		print "Sending as Photo: $fname\n";
		$result = eval { $api->sendPhoto(
			{
				chat_id => $Chat_ID,
				photo   => {
				    file => $fulfilename
				},
				disable_notification => $notifications,
				caption => $caption
			})
		} or $error_message = $api->parse_error->{msg};
	} elsif ( $fname =~ /\.gif$/i ) {
		writelog("Sending as Animation: $fname");
		print "Sending as Animation: $fname\n";
		$result = eval { $api->sendAnimation (
		    {
			chat_id => $Chat_ID,
			animation => {
			    file => $fulfilename
			},
			disable_notification => $notifications,
			caption => $caption
		    })
		} or $error_message = $api->parse_error->{msg};
	} elsif ( $fname =~ /\.mp4$/i ) {
		writelog("Sending as Video: $fname");
		print "Sending as Video: $fname\n";
		$result = eval { $api->sendVideo (
		    {
			chat_id => $Chat_ID,
			video   => {
			    file => $fulfilename
			},
			disable_notification => $notifications,
			caption => $caption
		    })
		} or $error_message = $api->parse_error->{msg};
	} else {
		print "Sending as Document: $fname\n";
		writelog("Sending as Document: $fname");
		$result = eval { $api->sendDocument (
		    {
			chat_id  => $Chat_ID,
			document => {
#				filename => $fulfilename
				file => $fulfilename
			},
			disable_notification => $notifications,
			caption => $caption
		    })
		} or $error_message = $api->parse_error->{msg};
	}
#	print " $error_message\n" if defined($error_message);
	print "SendDecoded Sleep ...\n\n";
	sleep 3;
}

sub tg_api_sendMessage()
{
    my $FromLen = length("$FromName $origAddr");
    my $ToLen = length($ToName);
    my $SubjLen = length($Subj);
    my $msgidlen = length($fidomsgid) if defined $fidomsgid;
#    my $TextLength = length( $Text );
    my ( $Chat_ID, %m_style );
    $notifications = 0;

    $m_style{'*'}='bold';
    $m_style{'/'}='italic';
    $m_style{'_'}='underline';
    my @t_entities = [{ offset => 2,
			length => 5,
			type => 'bold'
		    },{ offset => 8,
			length => ($FromLen),
			type => 'underline'
		    },{ offset => (13+$FromLen),
			length => 5,
			type => 'bold'
		    },{ offset => (19+$FromLen),
			length => $ToLen,
			type => 'underline'
		    },{ offset => (22+$FromLen+$ToLen),
			length => 5,
			type => 'bold'
		    },{ offset => (28+$FromLen+$ToLen),
			length => $SubjLen,
			type => 'underline'
		    },{ offset => (31+$FromLen+$ToLen+$SubjLen),
			length => $msgidlen,
			type => 'italic'
		    }];
    my ($t_url,$ut_offset,$ut_length);
    my $i = 6;
#    while ( $Text =~ /\[([^\[\]]+)\][\r\n\s]?\(([hf][t]{1,2}p[s]?\:\/\/[^\)]+)\)/i ||
#	    ( $Text =~ /([\_\*\/])[^\_\*\/\r\n]{1,76}([\_\*\/])/ && $1 eq $2 ) ) {
#	$i++;
#	if ( $1 ne $2 ) {
#	    $t_entities[0][$i]{type} ='text_link';
#	    $t_entities[0][$i]{url} = $2;
#	    $ut_offset = index($Text,"[$1]");
#	    $t_entities[0][$i]{offset} = $ut_offset;
#	    $t_entities[0][$i]{length} = index($Text,'](',$ut_offset) - $ut_offset - 1;
#	    writelog("[URL]: $2");
#	    $Text =~ s/\[([^\[\]]+)\][\r\n\s]?\(([hf][t]{1,2}p[s]?\:\/\/[^\)]+)\)/$1/i;
#	} elsif ( $1 eq $2 ) {
##	    $one = $1;
#	    if ( $Text =~ /([hf][t]{1,2}p[s]?\:\/\/[^\s\r\n]+)/i ) {
##		[\.\,\;\:\!]?
#		$t_entities[0][$i]{type} ='text_link';
#		$t_entities[0][$i]{url} = $1;
#		$t_entities[0][$i]{offset} = index($Text,$1);
#		writelog("URL: $1");
#		$t_entities[0][$i]{length} = 6;
#		$Text =~ s/([hf][t]{1,2}p[s]?\:\/\/[^\s\r\n]+)/\[LINK\]/i;
#	    } else {
#		$Text =~ /([\_\*\/])([^\_\*\/\r\n]{1,76})([\_\*\/])/;
#		$t_entities[0][$i]{type} = $m_style{$1};
#		$ut_offset = index($Text,"$1$2$3");
#		$t_entities[0][$i]{offset} = $ut_offset;
#		$t_entities[0][$i]{length} = length($2);
#		$Text =~ s/([\_\*\/])([^\_\*\/\r\n]{1,76})([\_\*\/])/$2/;
#	    }
#	}
#    }

    if ( $area ne 'NETMAIL' ) {
	$Chat_ID = $tags_id{$area};
	if ( defined( $options{$Chat_ID} ) ) {
	if ( $options{$Chat_ID} =~ /\-dn/i ) {
	    $notifications = 1;
	}
	}
    } else {
	$Chat_ID = $ToName;
    }
    undef $error_message;
    $result = eval { $api->sendMessage (
	   {
		chat_id  => $Chat_ID,
		text     => $Text,
		entities => @t_entities,
		disable_notification => $notifications,
		allow_sending_without_reply => 1,
		reply_to_message_id => $tg_msg_repl_to
	   }
    	)
    } or $error_message = $api->parse_error->{msg};
    print "sendMessage sleep...\n";
    sleep 3;
}

sub send_text_msg()
{
	my $cutlen;

	$Text=~s/[\r\n]SEEN\-BY\:[^\r\n]*/\r/ig;
#	$Text=~s/\x01Via/\@Via/ig if $FromName =~ /Ping/i;
	$Text=~s/[\r\n] \* Origin\:[^\r\n]*\)/\r/g;
	$Text=~s/\x01[^\r]*?\r//g;
	$Text=~s/\r\-\-\- [^\r]*\r/\r/g;
#	$Text =~ s/\+/\\\+/g;
	$Text =~ s/To\:\s*\d+\s\<\d+\@p?\d*\.?f\d+\.n\d+\.z\d+\.fidonet\.org\>\r//ig;
	$Text =~ s/\r/\n/g;
	my $mText = $Text;

	my $maxlen = 3850;
	my $ofset = 0;
	my $filelen = length($mText);
	while( $ofset < $filelen ){
	    $cutlen = index( $mText, ' ', $ofset+$maxlen-80 );
	    $cutlen = $maxlen if ( $cutlen == -1 || $cutlen == 0 || !defined($cutlen) );
	    $cutlen++;
	    $cutlen = $maxlen if $cutlen > $maxlen;
	    #print "Text Length: $filelen\nOffset: $ofset\nMaxLen: $maxlen\nCat Len: $cutlen\n\n";
            $Text = "--From: $FromName $origAddr--\n".
        	    "--To  : $ToName\n".
        	    "--Subj: $Subj--\n$fidomsgid\n".
	    substr( $mText, $ofset, $cutlen )."\n";
	    $ofset += $cutlen;
	    $Text = decode( 'cp866', $Text ) if $fromchrs ne 'utf-8';
	    $Text = decode( 'utf-8', $Text ) if $fromchrs eq 'utf-8';

	    tg_api_sendMessage();

	    if ( defined($error_message) ) {
		if ( $error_message =~ /replied message not found/i ) {
		    writelog("WARNING: ($tg_msg_repl_to) $error_message");
		    print "WARNING: ($tg_msg_repl_to) $error_message\n";
		    sleep 1;
		    undef $tg_msg_repl_to;
		    tg_api_sendMessage();
		}
		if ( $error_message =~ /bot can\'t send messages to bots/i ) {
#            $Text = "--From: $FromName $origAddr--\n".
#        	    "--To  : $ToName\n".
#        	    "--Subj: $Subj--\n$fidomsgid\n".
		    sendaspkt('Fido2Telebot', $FromName, $address, $origAddr,
			    "\rForbidden: You can\'t send messages to bots!".
			    "\r\r$tearline",
			    'Got Error!');
		    writelog($error_message);
		    undef $error_message;
		}
#		sleep 10;
	    }

	    if ( defined( $error_message ) ) {
		writelog("Got error: $error_message");
		print "Got error: \'$error_message\'\n";
	    }

#	    print "While Sleep ...\n\n";
#	    sleep 1;
	}
}

sub uu_filename($)
{
    ( $fname ) = @_;
#    print STDERR "$fname\n";
    $fulfilename = $temp_dir . $slash .
	    $area .	$slash . $fname;
#    print STDERR "$fulfilename\n";
    $fnu = url_encode( $fname );
    $fnu =~ s/\+/\%20/g; # fixing for urlencode
#   $filesize = -s $fulfilename;
    mkdir $temp_dir . $slash . $area unless -d $temp_dir . $slash . $area;
}

sub dec_uu()
{
    my ( @uuelines, $uudecoded_data,$UUOF );
    my $fo = 0;
    while ($Text =~ /\r\n?begin[ ]+\d+[ ]+([^\r]+)\r\n?([^ ]*?\r\n?)end\r\n?/) {
	$fo = 1;
	@uuelines = split(/\r\n?/,$2);
	uu_filename($1);

	if (open($UUOF, ">$fulfilename")){
	    binmode($UUOF);
	    foreach my $val ( @uuelines ){
	        $uudecoded_data = unpack("u", $val);
	        print($UUOF $uudecoded_data) if defined $uudecoded_data;
	    }
	    close($UUOF);
	    undef $UUOF;
	    undef @uuelines;
	} else {
	    writelog("Can't open \"$fulfilename\"\: $!\.");
	    print STDERR "Can't open \"$fulfilename\"\: $!\.\n";
    	}
	$filesize = ( stat($fulfilename) )[7];
#	print STDERR "\'$filesize\'\n";

#	10485760 for (jpg|png)
	if ( $fname =~ /(jpg|png)/i ) {
	    if( $filesize > 10485760 ){
	    $urlonly = 1;
	    undef $fileonly;
	    print "WARNING: Photo $fname ($filesize) is bigger 10Mb. Sending as URL.\n";
	    writelog("WARNING: Photo $fname ($filesize) is bigger 10Mb. Sending as URL.");
	    }
	}
	if ( $filesize  > 52428800 ) {
			# 50 Mb
	    $urlonly = 1;
	    undef $fileonly;
	    print "WARNING: $fname ($filesize) is bigger 50Mb. Sending as URL.\n";
	    writelog("WARNING: $fname ($filesize) is bigger 50Mb. Sending as URL.");
	}
    	
	writelog("      UUe detected. File name: $fname, Size: $filesize");
	print "      UUe detected. File name: $fname, Size: $filesize\n";
	
	$caption = decode( 'cp866', $Subj ) if $fromchrs ne 'utf-8';
	
	if ( !defined( $urlonly ) ) {
	    send_decoded($tags_id{$area}) if $area ne 'NETMAIL';
	    send_decoded($ToName) if $area eq 'NETMAIL';
	    if ( defined( $error_message ) ) {
		writelog("Got error: $error_message");
		print "Got error: \'$error_message\'\n";
	    }
	}
	
	print( STDERR "Can't delete $fulfilename! ($!).\n" ) && 
	   writelog( "Can't delete $fulfilename! ($!)." )
	   unless unlink $fulfilename;

	if ( defined ( $fileonly ) ) {
	    $Text =~ s/\r\n?begin[ ]+\d+[ ]+([^\r]+)\r\n?[^ ]*?\r\n?end\r\n?/\r/;
	} else {
	    if ( $area ne 'NETMAIL' ){
		$Text =~ s/\r\n?begin[ ]+\d+[ ]+([^\r]+)\r\n?[^ ]*?\r\n?end\r\n?/\r\rhttps\:\/\/brorabbit.g0x.ru\/uue\/$area\/$fnu \r(Size: $filesize)\r/;
	    } else {
		$Text =~ s/\r\n?begin[ ]+\d+[ ]+([^\r]+)\r\n?[^ ]*?\r\n?end\r\n?/\rFile sent abowe: $fname \r(Size: $filesize)\r/;
	    }
	}
    } # end while begin...end

#    undef $fileonly if $fo == 0;
return $fo;
}


sub dec_b64()
{
    my ( $encoded_data,$UUOF );
    my $fo = 0;
    while ($Text =~ /\r\n?Content\-type\: [^\r\n]+\; name\=\"([^\r\n\"]+)\"[^\r\n]*\r\n?Content\-transfer\-encoding\: base64\r\n?[\r\n]*([A-Z0-9\+\/\=\r\n]+)/i) {
	$fo = 1;
	$encoded_data = $2;
	uu_filename($1);
	$encoded_data =~ s/[\r\n]//g;
	writelog("      Base64 detected. File name: $fname, Size: $filesize");
	print "      Base64 detected. File name: $fname, Size: $filesize\n";

	if (open($UUOF, ">$fulfilename")){
	    binmode($UUOF);
	    print($UUOF decode_base64($encoded_data));
	    close($UUOF);
	    undef $UUOF;
	} else {
	    writelog("Can't open \"$fulfilename\"\: $!\.");
	    print STDERR "Can't open \"$fulfilename\"\: $!\.\n";
    	}
	$filesize = ( stat($fulfilename) )[7];
#	print STDERR "\'$filesize\'\n";

#	10485760 for (jpg|png)
	if ( $fname =~ /(jpg|png)/i ) {
	    if( $filesize > 10485760 ){
	    $urlonly = 1;
	    undef $fileonly;
	    print "WARNING: Photo $fname ($filesize) is bigger 10Mb. Sending as URL.\n";
	    writelog("WARNING: Photo $fname ($filesize) is bigger 10Mb. Sending as URL.");
	    }
	}
	if ( $filesize  > 52428800 ) {
			# 50 Mb
	    $urlonly = 1;
	    undef $fileonly;
	    print "WARNING: $fname ($filesize) is bigger 50Mb. Sending as URL.\n";
	    writelog("WARNING: $fname ($filesize) is bigger 50Mb. Sending as URL.");
	}
	
	$caption = decode( 'cp866', $Subj ) if $fromchrs ne 'utf-8';
	
	if ( !defined( $urlonly ) ) {
	    send_decoded($tags_id{$area}) if $area ne 'NETMAIL';
	    send_decoded($ToName) if $area eq 'NETMAIL';
	    if ( defined( $error_message ) ) {
		writelog("Got error: $error_message");
		print "Got error: \'$error_message\'\n";
	    }
	}
	
	print( STDERR "Can't delete $fulfilename! ($!).\n" ) && 
	   writelog( "Can't delete $fulfilename! ($!)." )
	   unless unlink $fulfilename;

	if ( defined ( $fileonly ) ) {
	    $Text =~ s/\r\n?Content\-type\: [^\r\n]+\; name\=\"([^\r\n\"]+)\"[^\r\n]*\r\n?Content\-transfer\-encoding\: base64\r\n?[\r\n]*([A-Z0-9\+\/\=\r\n]+)/\r/i;
	} else {
	    if ( $area ne 'NETMAIL' ){
		$Text =~ s/\r\n?Content\-type\: [^\r\n]+\; name\=\"([^\r\n\"]+)\"[^\r\n]*\r\n?Content\-transfer\-encoding\: base64\r\n?[\r\n]*([A-Z0-9\+\/\=\r\n]+)/\r\rhttps\:\/\/brorabbit.g0x.ru\/uue\/$area\/$fnu \r(Size: $filesize)\r/i;
	    } else {
		$Text =~ s/\r\n?Content\-type\: [^\r\n]+\; name\=\"([^\r\n\"]+)\"[^\r\n]*\r\n?Content\-transfer\-encoding\: base64\r\n?[\r\n]*([A-Z0-9\+\/\=\r\n]+)/\rFile sent abowe: $fname \r(Size: $filesize)\r/i;
	    }
	}
    } # end while begin...end

#    undef $fileonly if $fo == 0;
return $fo;
}

sub namesfordupes()
{
    if (defined($area)){
	$echo_tag = $area;
    } else {
	$echo_tag = 'netmail';
    }
    $dupe_file = "${dupe_dir}${slash}${echo_tag}.dupe";
}

sub read_dupes()
{
    my ( $line );

    return if defined( $dupe_msgids{$echo_tag}{0} );
    
    if( open( $DFH, '<', $dupe_file ) ) {
	writelog("Reading dupes for \'$echo_tag\'.");
	$dupe_msgids{$echo_tag}{0} = 1;
	while ( $line = <$DFH> ){
	    $line =~ s/[\r\n]//g;
	    $line =~ /^(\d+) (.*)$/;
	    $dupe_msgids{$echo_tag}{$2} = 1;
	}
	close( $DFH );
	writelog("Done.");
    } else {
	writelog( "ERROR: Can't open \'$dupe_file\' to read ($!)." );
	print "ERROR: Can't open \'$dupe_file\' to read ($!).\n";
    }
}

#our %moder;

#sub moderate()
#{
#  return if defined($area)
#  return if $destAddr ne $address
##    writelog("From: $origAddr");
## $FromName
## $ToName
#
#	if ($Text =~ /\%area ([^\r\n\s]+)/i) {
#	    if( defined( $moder{$area}{modpswd} ) ){
#	    }
#	}
#}

sub readPKT($)
{
	my ( $pktfilename ) = @_;
	my ( $filefound );
#	print "$pktfilename\n";
	if ( open(PKTF, "<$pktfilename") ){

		if ( read( PKTF, $pkthdr, 0x3a ) != 0x3a ) {
		    close(PKTF);
		    $pktfilename =~ /(.*)pkt$/i;
		    $badname = $1 . 'bad';
		    print( STDERR "Broken packet - invalid header size.\n" );
		    writelog("Broken packet - invalid header size.");
		    writelog("   Renaming to $badname");
		    move( $pktfilename, $badname);
		    return 0;
		}
		
		# --- partial unused now ---------
		my ($origNode,$destNode,$year,$month,$day,$hour,$minute,
			$seconds,$baud,$type,$origNet,$destNet,
			     # Follows Type2+ packet fields...
			$ProductCode,$RevMaj,$Password,$QMOrigZone,$QMDestZone,
			$AuxNet,$CapValidate,$PCodeHi,$RevMin,$Cap,$origZone,
			$destZone,$origPoint,$destPoint,$appdata)=
			unpack('S2S3S3 S2S2 C2A8S2S2 C2SS4I',$pkthdr);
		# --- partial unused now ---------
		print "   from $origZone:$origNet/$origNode.$origPoint to $destZone:$destNet/$destNode.$destPoint\n";
		writelog("   from $origZone:$origNet/$origNode.$origPoint to $destZone:$destNet/$destNode.$destPoint");

		if ( "$destZone:$destNet/$destNode.$destPoint" ne $VirtualAddress ) {
			print STDERR "   PKT not for us!\n";
			writelog("   PKT not for us!");
			close(PKTF);
			$pktfilename =~ /(.*)pkt$/i;
			$badname = $1 . 'ntu';
			print STDERR "   Renaming to $badname\n";
			writelog("   Renaming to $badname");
			move( $pktfilename, $badname );
			return 0;
		}
		if ( "$origZone:$origNet/$origNode.$origPoint" ne $uplink) {
			print STDERR "ERROR: PKT not from uplink.\n";
			writelog("ERROR: PKT not from uplink.");
#			close(PKTF);
		}
		if( defined( $pktpassword ) && $Password ne $pktpassword ) {
			close(PKTF);
			print STDERR "   Wrong PKT password.\n";
			writelog("   Wrong PKT password.");
			$pktfilename =~ /(.*)pkt$/i;
			$badname = $1 . 'sec';
			print STDERR "   Renaming to $badname\n";
			writelog("   Renaming to $badname");
			move( $pktfilename, $badname );
			return 0;
		}
# --- reading messages ---
		undef $Text;
		$oldcr = $/;
		while(read(PKTF,$version,2)==2 &&
				($pktver=unpack('S',$version))==2) {

			if(read(PKTF,$hdr,12)!=12) {
				close(PKTF);
				$pktfilename =~ /(.*)pkt$/i;
				$badname = $1 . 'bad';
				print( STDERR "Broken packet - invalid message header.\n" );
				writelog("Broken packet - invalid message header.");
				writelog("   Renaming to $badname");
				move( $pktfilename, $badname );
				return 0;
			}

			($origNode,$destNode,$origNet,$destNet,$attr,$cost)=unpack('S6',$hdr);
			$/="\0";
			$dateTime=<PKTF>;
			chop $dateTime;
			$ToName=<PKTF>;
			chop $ToName;
			$FromName=<PKTF>;
			chop $FromName;
			$Subj=<PKTF>;
			chop $Subj;
			$Text=<PKTF>;
			chop $Text;
			$/=$oldcr;
			if(length($dateTime)>19||length($ToName)>35||
				length($FromName)>35||length($Subj)>71) {
				print STDERR "Warning: Bad field(s) length (too long)\n";
				writelog( "Warning: Bad field(s) length (too long)!" );
			}
			$msg_crc = crc32($Text);
			$Text.="\r" unless($Text=~/\r\n?$/s);
			undef $area;
			if($Text=~/^AREA:\s*(\S+)\r\n?/m) {
				$area=uc($1);
				undef $area if($area eq 'NETMAIL');
				substr($Text,length($`),length($&))='';
			}
			undef $fromchrs;
			if ( $Text =~ /\001CHRS\:\s*(\S+)\s+\d+/ ) {
#			print ("$area $1 $2\n");
				writelog("\'$area\' $1") if defined $area;
				$fromchrs = lc($1);
				writelog("Charset found: \'$fromchrs\'.") if !defined $area || defined $tags_id{$area};
			} else {
				$fromchrs = 'cp866';
				writelog('Warning: No @CHRS kludge. Using default CP866.') if !defined $area || defined $tags_id{$area};
			}
			if($area) {
				undef $origAddr;
				undef $destAddr;
				next unless defined $tags_id{$area};
			} else {
				# is NETMAIL
				next unless $ToName =~ /^\d+$/;
				$origAddr="$origNet/$origNode";
				$destAddr="$destNet/$destNode";
				if($Text=~/^\x01INTL:?[ ]+(\d+:\d+\/\d+)[ ]+(\d+:\d+\/\d+)[ ]*\r\n?/m) {
					$destAddr=$1;
					$origAddr=$2;
					substr($Text,length($`),length($&))='';
				}
				if($Text=~/^\x01FMPT:?[ ]+(\d+)[ ]*\r\n?/m) {
					$origAddr.=".$1";
					substr($Text,length($`),length($&))='';
				}
				if($Text=~/^\x01TOPT:?[ ]+(\d+)[ ]*\r\n?/m) {
					$destAddr.=".$1";
					substr($Text,length($`),length($&))='';
				}
				writelog("Netmail detected.");
				writelog("From: $origAddr");
				writelog("To  : $destAddr");
			}
			if(!$origAddr){
			if( $Text =~ /\r\n? \* Origin:[^\r]*\((\d+:\d+\/\d+.?\d*)\)\r\n?/m ) {
				$origAddr=$1;
			}}
			undef $msgid_addr;
			undef $fidomsgid;
			if($Text=~/\x01MSGID:[ ]*(\S+)[ ]+(\S+)[ ]*\r\n?/i) {
				$fidomsgid = "$1 $2";
				$msgid_addr=$1;
				if ( $msgid_addr =~ /(\d+\:\d+\/\d+\.?\d*)/ ) {
					$origAddr=$1;
				}
#				if (defined($area)){
#				    $dupe_file = "${dupe_dir}${slash}${area}.dupe";
#				} else {
#				    $dupe_file = "${dupe_dir}${slash}netmail.dupe";
#				}
#				my $area_name;
#				if ( defined($area) ) {
#				    $area_name = $area;
#				} else {
#				    
#				}
			} else {
				writelog( 'WARNING: @MSGID not found.' );
				print "WARNING: \@MSGID not found.\n";
			}

			$tg_msg_repl_to = undef;
			if($Text=~/\x01REPLY\:[ ]*$address[ ]+([a-f\d]{8})\r\n?/i) {
				$tg_msg_repl_to = hex($1);
				print "      Telegram replay found \'$1\'.\n";
				writelog("      Telegram replay found \'$1\'.");
			}

			print "      Area: $area ( $tags_id{$area} )\n" if defined $area;
			print("      From: $FromName $origAddr\n");
			print("      To  : $ToName\n");
			print("      Subj: $Subj\n");
			print("      ------------------------------------------------\n");
			writelog("Area: $area ( $tags_id{$area} ) options: \'$options{$area}\'") if defined $area;
			writelog("From: $FromName $origAddr");
			writelog("To  : $ToName");

			undef $urlonly;
			undef $fileonly;
			if ( defined $area ){
			    $urlonly = 1 if ( $options{$area} =~ /urlonly/i );
			    $fileonly = 1 if ( $options{$area} =~ /fileonly/i );
			} else {
			    $area='NETMAIL';
			}
# -- UUDECODE -->			
			$filefound = dec_uu();
			$filefound += dec_b64();
			$Text =~ s/begin[ ]+\d+[ ]+([^\r\n]+)[\r\n]\n?.*[\r\n]\n?end[\r\n]\n?//gsi;
			undef $fileonly if $filefound == 0;
# -- UUDECODE end -->			

			$fidomsgid = '2:460/256 ' . sprintf( "%08x", time() )
			    unless defined( $fidomsgid ) && defined( $area );

			if ( defined($fidomsgid) ) {
			    namesfordupes();
			    read_dupes();
			    if ( defined( $dupe_msgids{$echo_tag}{$fidomsgid} ) ) {
				writelog("Warning! Duplicate message found ($dateTime $fidomsgid).");
				writelog("  Area: $echo_tag");
				writelog("  From: $FromName $origAddr");
				writelog("  To  : $ToName");
				next;
			    }
			}
			send_text_msg() unless defined $fileonly;
			close(PKTF) && return 0 if defined($error_message);
			if ( defined($fidomsgid) ) {
			    if( open( $DFH, '>>', $dupe_file ) ) {
			        $dupe_msgids{$echo_tag}{$fidomsgid} = 1;
			        print $DFH time()." $fidomsgid\n";
			        close( $DFH );
			    } else {
				writelog( "ERROR: Can't open \'$dupe_file\' to write ($!)." );
				print "ERROR: Can't open \'$dupe_file\' to write ($!).\n";
			    }
			}
			# --- reading messages
		} # --- while read PKT end ---

		close(PKTF);

		if (!unlink( $pktfilename )) {
		    print( STDERR "Can't delete $pktfilename. ($!).\n" );
		    writelog( "Can't delete $pktfilename. ($!)." );
		}

	} else {
		print(STDERR "Can't open $pktfilename. ($!).\n");
		writelog("Can't open $pktfilename. ($!).");
		return 0;
	}
return 1;
}

sub gone_or_new_users()
{
	my ( $goneuserfirst, $goneuserlast, $newusercount, $newuserfirst,
	$newuserlast, $newuser_username );

	$goneuserfirst = ( $api->getUpdates->{result}[$u]{message}->{left_chat_member}->{first_name} );
	$goneuserlast  = ( $api->getUpdates->{result}[$u]{message}->{left_chat_member}->{last_name} );

	$msgtext = $goneuserfirst if defined $goneuserfirst;
	$msgtext .= " $goneuserlast" if defined $goneuserlast;
	$msgtext .= ' left the chat.' if defined $goneuserfirst;;
	
	if( defined( ($api->getUpdates->{result}[$u]{message}->{new_chat_members}) ) ) {
	    $newusercount = @{ ($api->getUpdates->{result}[$u]{message}->{new_chat_members}) };
	    print "newusercount: \'$newusercount\'\n";
	    for ($nu = 0; $nu < $newusercount; $nu++) {
		$newuserfirst = ( $api->getUpdates->{result}[$u]{message}->{new_chat_members}[$nu]->{first_name} );
		$newuserlast  = ( $api->getUpdates->{result}[$u]{message}->{new_chat_members}[$nu]->{last_name} );
		$newuser_username = ( $api->getUpdates->{result}[$u]{message}->{new_chat_members}[$nu]->{username} );
		$msgtext .= $newuserfirst;
		$msgtext .= " $newuserlast" if defined $newuserlast;
		$msgtext .= " \(User name: $newuser_username\)" if defined $newuser_username;
		$msgtext .= " joined the group.\r\n";
	    }
	}

    $subj = 'New user' if defined $newuserfirst;
    $subj = 'User gone' if defined $goneuserfirst;
}

sub get_from_names()
{
#use locale;
print "Getting from names.\n";
    my ( $first_from_name, $last_from_name, $fwd_first_from_name,
	$fwd_last_from_name );

    undef $fwd_fromname;
    undef $fromname;

    $first_from_name = ( $api->getUpdates->{result}[$u]{message}{from}->{first_name} );
    $last_from_name = ( $api->getUpdates->{result}[$u]{message}{from}->{last_name} );
    $from_user_id = ( $api->getUpdates->{result}[$u]{message}{from}->{id} );
    $fromname = $first_from_name if defined $first_from_name;
    $fromname .= " $last_from_name" if defined $last_from_name;
    $fromname = encode('cp866', $fromname);
    
    $fwd_first_from_name = ( $api->getUpdates->{result}[$u]{message}{forward_from}->{first_name} );
    $fwd_last_from_name = ( $api->getUpdates->{result}[$u]{message}{forward_from}->{last_name} );
    $fwd_fromname = $fwd_first_from_name if defined $fwd_first_from_name;
    $fwd_fromname .= " $fwd_last_from_name" if defined $fwd_last_from_name;
    $fwd_fromname = ( $api->getUpdates->{result}[$u]{message}->{forward_sender_name} ) if !defined $fwd_fromname;
    $fwd_fromname = encode('cp866', $fwd_fromname) if defined $fwd_fromname;

    $fromname =~ s/[\r\n\000]//g;
    $fromname = substr( $fromname, 0, 35 ) if length( $fromname ) >35;
print "From Name: $fromname\n";
print "Done.\n";
}

sub get_to_name()
{
#use locale;
print "Getting To name.\n";
	$toname = 'All';
	$toaddr = '2:460/58';
	undef $fidomsgid;
	my ( $first_toname, $last_toname, @qc );
	
	if ( defined($replytext) ) {
		$first_toname = ( $api->getUpdates->{result}[$u]{message}{reply_to_message}{from}->{first_name} );
		$last_toname =  ( $api->getUpdates->{result}[$u]{message}{reply_to_message}{from}->{last_name} );
		$to_user_id =  ( $api->getUpdates->{result}[$u]{message}{reply_to_message}{from}->{id} );
		$toname = encode( 'cp866', $first_toname) if defined $first_toname;
		$toname .= encode( 'cp866', " $last_toname") if defined $last_toname;
		
		if ( "$replytext" =~ /^[\r\n]*\-\-From: ([^\r\n]+) (\d+\:\d+\/\d+\.?\d*)\-\-[\r\n]+\-\-To  : [^\r\n]+[\r\n]([^\r\n]+)[\r\n]+/i ) {
		    ( $toname, $toaddr, $fidomsgid ) = ( $1, $2, $3 );
#		    $replytext =~ s/\-\-To  : ([^\r\n]+)[\r\n]+//i;
		    $replytext =~ s/^[\r\n]*\-\-From: ([^\r\n]+) (\d+\:\d+\/\d+\.?\d*)\-\-[\r\n]+\-\-To  : [^\r\n]+[\r\n]+([^\r\n]+)[\r\n]+//i;
#---> Fido @MSGID
#		    undef $reply_msgid;
		    $reply_msgid = $fidomsgid;
		}
	    $quotechars = $toname;
	    $quotechars =~ s/_/ /g;
	    $quotechars =~ s/\./ /g;
	    @qc = split(' ', $quotechars);
	    $quotechars = '';
	    for my $ww (@qc){
		$quotechars .= substr($ww,0,1);
	    }
	}
	if ( defined( $msgtext ) ) {
		if ( $msgtext =~ /^[\r\n]*([\-\.\=\,\+\/\$])\1To\:\s*([^\r\n]+?)\s?(\d+\:\d+\/\d+\.?\d*)?\1\1\r?\n?/i ) {
			$toname = $2;
			$toaddr = $3 if $3; #for netmail
			$msgtext =~ s/^[\r\n]*([\-\.\=\,\+\/\$])\1To\:\s*([^\r\n]+?)\s?(\d+\:\d+\/\d+\.?\d*)?\1\1\r?\n?//i;
		}
	}
    $toname =~ s/[\r\n\000]//g;

    $toname = substr( $toname, 0, 35 ) if length( $toname ) > 35;
print "To Name: $toname\n";
print "To Address: $toaddr\n" if defined $toaddr;
print "Done\n";
}

sub set_subj()
{
#    return if defined $subj;
    print "Setting Subj...\n";
    undef $subj;
    if ( defined($replytext) ) {
	if ( $replytext =~ /[\-\.\=\,\+\/]{2}Subj:\s*([^\r\n]*?)[\-\.\=\,\+\/]{2}\r/i ){
	    $subj = $1;
	    $replytext =~ s/[\-\.\=\,\+\/]{2}Subj:\s*([^\r\n]*?)[\-\.\=\,\+\/]{2}\r//i;
	}
    }
    if ( defined( $msgtext ) ) {
	if ( $msgtext =~ /[\-\.\=\,\+\$\/]{2}Subj:\s*([^\r\n]*)[\-\.\=\,\+\/\$]{2}\r/i ){
	    $subj = $1;
	    $msgtext =~ s/[\-\.\=\,\+\/\$]{2}Subj:\s*([^\r\n]*)[\-\.\=\,\+\/\$]{2}\r//i;
	}elsif ( !defined( $subj ) ) {
	    if ( $msgtext =~ /^[\r\n]*([^\r\n]{1,69}[\r\n\s\.\,\-\~\=\!\"\'\\\/\?])/ ) {
		$subj = $1;
	    }
	}
    }
    $subj = '...' if !defined( $subj );
    $subj = substr( $subj, 0, 71 ) if defined($subj) && length( $subj ) > 71;
    $subj =~ s/[\r\n\000]//g;
    print "Subj: $subj\n";
    print "Done.\n";
}

sub sendaspkt($;$$$$$$)
{
    print "Sending as PKT...\n";
    writelog("Sending as PKT...");
    my ($msg_fromname, $msg_toname, $msg_fromaddr, $msg_toaddr, $msg_txt,
	$msg_subj, $msg_area ) = @_;
    
    $msg_toaddr = $address if !defined $msg_toaddr;
    writelog("Area: $msg_area") if defined $msg_area;
    writelog("From: $msg_fromname $msg_fromaddr");
    writelog("To  : $msg_toname $msg_toaddr");
    writelog("Subj: $msg_subj");
#    writelog("\'$msg_txt\'");
    
    if (length($msg_fromname) > 35){
	$msg_fromname = substr($msg_fromname,0,35);
    }
    $msg_fromname .= "\000";
    if (length($msg_toname) > 35){
	$msg_toname = substr($msg_toname,0,35);
    }
    $msg_toname .= "\000";
    if (length($msg_subj) > 71){
	$msg_subj = substr($msg_subj,0,71);
    }
    $msg_subj .= "\000";

    $VirtualAddress =~ /^(\d+)\:(\d+)\/(\d+)\.?(\d*)$/;
    my ( $pkt_origzone, $pkt_orignet, $pkt_orignode, $pkt_origpnt ) = ( $1, $2, $3, $4 );
    $pkt_origpnt = 0 if !defined $pkt_origpnt;
    $uplink =~ /^(\d+)\:(\d+)\/(\d+)\.?(\d*)$/;
    my ( $pkt_destzone,$pkt_destnet,$pkt_destnode,$pkt_destpnt ) = ( $1,$2,$3,$4 );
    $pkt_destpnt = 0 if !defined $pkt_destpnt;
    my $password = $pktpassword;
    
    localtime =~ /[a-z]+ ([a-z]+)[ ]+(\d+) (\d+)\:(\d+)\:(\d+) \d\d(\d\d)/i;
    my $DateTime = sprintf("%02s", $2)." $1 $6  $3:$4:$5\000";

    my ($second,$minute,$hour,$day,$month,$year,$wday,$yday,$isdst) = localtime();
    $year = $year + 1900;
    $yday++;
    my $pktname = $outpktdir . $slash . sprintf("%08x", time()).'.pkt';
    while( -e $pktname ) {
	sleep 1;
	$pktname = $outpktdir . $slash . sprintf("%08x", time()).'.pkt';
    }

    my ($pkthdr,$msgheader,$m_txt);

    $msg_fromaddr =~ /(\d+)\:(\d+)\/(\d+)\.?(\d*)/;
    my ( $msg_origzone, $msg_orignet, $msg_orignode, $msg_origpnt ) = ( $1, $2, $3, $4 );
    $msg_origpnt = 0 if !defined( $msg_origpnt ) || $msg_origpnt eq "";

    $msg_toaddr =~ /(\d+)\:(\d+)\/(\d+)\.?(\d*)/;
    my ($msg_destzone,$msg_destnet,$msg_destnode,$msg_destpnt) = ( $1, $2, $3, $4 );
    $msg_destpnt = 0 if !defined($msg_destpnt) || $msg_destpnt eq '';

    $pkthdr = pack("S2S3S3 S2S2 C2A8S2S2 C2SS4I", $pkt_orignode,
    $pkt_destnode,$year,$month,
    $day,$hour,$minute,$second,0,2,$pkt_orignet,$pkt_destnet,3,3,
    $password,$pkt_origzone,$pkt_destzone, 
    0, 0x0100,3,3,0x0001,
    $pkt_origzone,$pkt_destzone,$pkt_origpnt,$pkt_destpnt, 0 );

    $msgheader = pack( "S7Z20", 2,$msg_orignode,$msg_destnode,
    $msg_orignet,$msg_destnet,0,0,$DateTime ) . $msg_toname .
    $msg_fromname . $msg_subj;

    $m_txt = "AREA:$msg_area\r" if defined($msg_area);
    if(defined($tg_msg_id)){
	$m_txt .= sprintf( "\001MSGID: $msg_fromaddr %08x\r", $tg_msg_id );
    } else {
	$m_txt .= "\001MSGID: $msg_fromaddr ".sprintf("%08x", time())."\r";
    }
    
#    $m_txt .= sprintf ("\001REPLY: $msg_fromaddr %08x\r", $reply_msgid ) if defined $reply_msgid;
    $m_txt .= "\001REPLY: $reply_msgid\r" if defined $reply_msgid;
    
# NO INTL in Echmail permitted
    $m_txt .= "\001INTL $msg_destzone:$msg_destnet/$msg_destnode $msg_origzone:$msg_orignet/$msg_orignode\r" if !defined $msg_area;
    $m_txt .= "\001PID: tg_BBS_v$vers\r\x01TZUTC: $TZUTC\r";
    $m_txt .= "\001FMPT $msg_origpnt\r" if $msg_origpnt != 0;
    $m_txt .= "\001TOPT $msg_destpnt\r" if $msg_destpnt != 0 && !defined $msg_area;

    $msg_txt .= " \* Origin: $origin \($msg_fromaddr\)\r";

    my ( $g_sec, $g_min, $g_hour, $g_mday,$g_month,$g_year) = (gmtime)[0...5];
    if ( defined( $msg_area ) ) {
	if ( $pkt_origpnt == 0 ) {
	    $msg_txt .= "SEEN\-BY: $msg_orignet\/$msg_orignode $pkt_destnet\/$pkt_destnode\r".
			"\x01PATH: $msg_orignet\/$msg_orignode\r";
	}
    } else {
	$msg_txt .= sprintf("\x01Via $address \@%04d%02d%02d\.%02d%02d%02d\.UTC tg_BBS $vers\r",
	$g_year+1900, $g_month+1, $g_mday, $g_hour, $g_min, $g_sec );
    }
    $msg_txt .= "\000\000\000";

    if ( open(FF, ">$pktname") ) {
	print(FF "$pkthdr$msgheader$m_txt$msg_txt");
	close(FF);
	open( FLAG, ">/home/fido/flags/toss.now" ) && close( FLAG );
    } else {
	print( STDERR "Can't create $pktname: $!\n");
	writelog("Can't create $pktname: $!");
##	return "Can't create $pktname: ($!)";
    }
print "Done.\n";
writelog("Sending as PKT done.");
}

sub get_entities()
{
    my ( $meCount, $meType, $stroff, $resstr, %style, $strlen, $c, $offset,
    $length, $e_url, $e_str );
    return if !defined( ( $api->getUpdates->{result}[$u]{message}->{entities} ) );
    $style{bold} = '*';
    $style{italic} = '/';
    $style{underline} = '_';
    $meCount = @{( $api->getUpdates->{result}[$u]{message}->{entities} )};
    $stroff = 0;
    $resstr = '';
    $strlen = 0;
    for( my $c=0; $c < $meCount; $c++ ) {
         $meType = ( $api->getUpdates->{result}[$u]{message}{entities}[$c]->{type} );
         if ( $meType =~ /(bold)|(italic)|(underline)/ ) {
#             $meType = ( $api->getUpdates->{result}[$u]{message}{entities}[$c]->{type} );
             $offset = ( $api->getUpdates->{result}[$u]{message}{entities}[$c]->{offset} );
             $length = ( $api->getUpdates->{result}[$u]{message}{entities}[$c]->{length} );
             $strlen = $offset - $stroff;
             $resstr .= substr( $msgtext, $stroff, $strlen );
             $resstr .= $style{$meType};
             $stroff += $strlen;
             $e_str = substr( $msgtext, $stroff, $length );
#             if ( $e_str =~ /^(.*?)[^\r\n\s]+$/) {
#                $e_str = $1;
		$e_str =~ s/^(.+)[\r\n\s]+$/$1/g;
                $length = length($e_str);
#             }
             $resstr .= $e_str;
             $resstr .= $style{$meType};
             $stroff += $length;
         }elsif ( $meType =~ /text_link/ ) {
             $offset = ( $api->getUpdates->{result}[$u]{message}{entities}[$c]->{offset} );
             $length = ( $api->getUpdates->{result}[$u]{message}{entities}[$c]->{length} );
             $e_url  = ( $api->getUpdates->{result}[$u]{message}{entities}[$c]->{url} );
             $strlen = $offset - $stroff;
             $resstr .= substr( $msgtext, $stroff, $strlen );
             $resstr .= '[';
             $stroff += $strlen;
             $resstr .= substr( $msgtext, $stroff, $length );
             $resstr .= ']';
             $stroff += $length;
             $resstr .= "\($e_url\)";
         }
    }
    $resstr .= substr( $msgtext, $stroff ) if $stroff > 0;
    $msgtext = $resstr if $resstr ne '';
    return;
#     {reply_to_message}
    return if !defined $replytext;
    return if !defined( ( $api->getUpdates->{result}[$u]{message}{reply_to_message}->{entities} ) );
    $meCount = @{( $api->getUpdates->{result}[$u]{message}{reply_to_message}->{entities} )};
    $stroff = 0;
    $resstr = '';
    $strlen = 0;
    for( my $c=0; $c < $meCount; $c++ ) {
         $meType = ( $api->getUpdates->{result}[$u]{message}{reply_to_message}{entities}[$c]->{type} );
         if ( $meType =~ /(bold)|(italic)|(underline)/ ) {
#             $meType = ( $api->getUpdates->{result}[$u]{message}{entities}[$c]->{type} );
             $offset = ( $api->getUpdates->{result}[$u]{message}{reply_to_message}{entities}[$c]->{offset} );
             $length = ( $api->getUpdates->{result}[$u]{message}{reply_to_message}{entities}[$c]->{length} );
             $strlen = $offset - $stroff;
             $resstr .= substr( $replytext, $stroff, $strlen );
             $resstr .= $style{$meType};
             $stroff += $strlen;
             $resstr .= substr( $replytext, $stroff, $length );
             $resstr .= $style{$meType};
             $stroff += $length;
         }
    }
     $resstr .= substr( $replytext, $stroff ) if $stroff > 0;
     $replytext = $resstr if $resstr ne '';
}

sub get_file()
{
    print "Getting file...\n";
    my ( $file_name, $photo_count, $file_path, $file_size, $file_id,
	$local_fname, $caption, $area_name );
    undef $error_message;
    undef $local_fname;
    $msgtext = '' if !defined $msgtext;

    $file_id = ( $api->getUpdates->{result}[$u]{message}{document}->{file_id} );
    $file_id = ( $api->getUpdates->{result}[$u]{message}{animation}->{file_id} ) if !defined($file_id);
    if ( defined($file_id) ){
	$file_name = ( $api->getUpdates->{result}[$u]{message}{document}->{file_name} );
	$file_name = ( $api->getUpdates->{result}[$u]{message}{animation}->{file_name} ) if !defined $file_name;
    } else {
	$photo_count = @{( $api->getUpdates->{result}[$u]{message}->{photo})}
		if defined $api->getUpdates->{result}[$u]{message}->{photo};
        if ( defined($photo_count) ) {
	    $photo_count--;
	    $file_id = ( $api->getUpdates->{result}[$u]{message}{photo}[$photo_count]->{file_id} );
	}
        $file_id = ( $api->getUpdates->{result}[$u]{message}{audio}->{file_id} ) if !defined($file_id);
        $file_id = ( $api->getUpdates->{result}[$u]{message}{video}->{file_id} ) if !defined($file_id);
        $file_id = ( $api->getUpdates->{result}[$u]{message}{video_note}->{file_id} ) if !defined($file_id);
        $file_id = ( $api->getUpdates->{result}[$u]{message}{voice}->{file_id} ) if !defined($file_id);
        $file_id = ( $api->getUpdates->{result}[$u]{message}{sticker}->{file_id} ) if !defined($file_id);
        $file_name = sprintf("%08x", time());
    }
    if ( defined($file_id) ) {
	$area_name = $areaname{$chatid};
	$area_name = 'NetMail' unless defined $area_name;
	writelog("From: $fromname, Area: $area_name, ".encode('cp866',$file_name).".");
	$file_size = eval{ (($api->api_request ( 'getFile',{file_id => "$file_id"}))->{result}->{file_size})
			} or $error_message = $api->parse_error->{msg};
	$file_path = eval{ (($api->api_request ( 'getFile',{file_id => "$file_id"}))->{result}->{file_path})
			} or $error_message = $api->parse_error->{msg};
	if ( defined( $error_message ) ){
	    writelog( "ERROR: getFile got \'$error_message\'" );
	    print STDERR "ERROR: getFile got \'$error_message\'\n";
#		rm_lock();
#		exit;
	    $caption = ( $api->getUpdates->{result}[$u]{message}->{caption} );
	    $msgtext .= encode('cp866',$caption) if defined $caption;
	    $msgtext .= "\r File: ".encode('cp866',$file_name).".\rError: $error_message";
	    return;
	}
	if ( defined( $file_size ) && $file_size > 20971520 ){
	    print("Your file was so big.\nIt might be very useful.\nBut now it is gone.\n");
	    writelog("ERROR: The file size is more than 20MB. The file could not be received.");
	    return;
	}
	if ($file_name !~ /\./){
	    $file_path =~ /.*[\\\/][^\\\/]+(\.[^\\\/]+)$/i;
	    $file_name .= $1;
	}
	print "file: $file_path\nFile Name: $file_name\nsize: $file_size\n";
	writelog("file: $file_path. File Name: $file_name. size: $file_size");
#	$get_file_url = "$get_file_url$bot_token/$file_path";
        $file_buffer = get("$get_file_url$bot_token/$file_path");
        if (defined($file_buffer)) {
	    $local_fname = $web_files_path . $slash . encode('cp866',$file_name);
	    if (open(OFH, ">$local_fname")) {
		binmode(OFH);
		print OFH $file_buffer;
		close(OFH);
	    } else {
		writelog("Can't open for write \'$local_fname\'. ($!)");
		print STDERR "Can't open to write \'$local_fname\'. ($!)";
	    }
#--->
	    $area_name = $areaname{$chatid};
	    if ( !defined( $area_name ) ) {
		$msgtext .= "\r".uu_encode( $local_fname, 644 )."\r";
		unlink $local_fname or warn "Could not unlink \'$local_fname\': $!" && writelog("Could not unlink \'$local_fname\': $!");
	    }elsif ( $options{$area_name} =~ /toUUE/i ) {
		$msgtext .= "\r".uu_encode( $local_fname, 644 )."\r";
		unlink $local_fname or warn "Could not unlink \'$local_fname\': $!" && writelog("Could not unlink \'$local_fname\': $!");
	    }elsif( defined($local_url) ){
		$file_name = url_encode( encode('cp866', $file_name) );
#		$file_buffer = url_encode( $file_name ); # put encoded ulr here
		$file_name =~ s/\+/%20/g; #fix error in url_encode
		$msgtext .= "\r\n$local_url$file_name\r\n";
	    } else {
		$msgtext .= uu_encode( $local_fname, 644 );
		unlink $local_fname or warn "Could not unlink \'$local_fname\': $!" && writelog("Could not unlink \'$local_fname\': $!");
	    }
	} else {
	    writelog("Can't get \'$get_file_url$bot_token/$file_path\'.");
	    print STDERR "Can't get \'$get_file_url$bot_token/$file_path\'.";
        }
    }
print "Done.\n";
}

sub readuserprofile($)
{
    my ( $u_id ) = @_;
    my $UsPrFile = File::Spec->catfile( $UsersProfileDir, "$u_id.conf" );

		    if ( open( UPFH, "<$UsPrFile" ) ) {
			$UserProfile{$u_id}{0} = 1;
			while( my $line = <UPFH> ) {
#			    $line = decode( 'cp866', $line );
			    $line =~ s/[\r\n]//g;
			    if( $line =~ /^\%?RealName\:? (.+)$/ ) {
				$UserProfile{$u_id}{realname} = $1;
			    } elsif ( $line =~ /^\%?TagLine\:? (.+)$/ ) {
				$UserProfile{$u_id}{tagline} = $1;
			    } elsif ( $line =~ /^\%?FidoAddr\:? (.+)$/ ) {
				$UserProfile{$u_id}{fidoaddr} = $1;
			    } elsif ( $line =~ /^\%?Greeting\:? (.+)$/ ) {
				$UserProfile{$u_id}{greeting} = $1;
			    } elsif ( $line =~ /^\%?Conclusion\:? (.+)$/ ) {
				$UserProfile{$u_id}{conclusion} = $1;
			    } elsif ( $line =~ /^\%?Tearline (.+)$/ ) {
				$UserProfile{$u_id}{tearline} = $1;
			    }
			}
			close( UPFH );
			$UserProfile{$u_id}{0} = 0 unless defined ( $UserProfile{$u_id}{realname} );
		    } else {
			writelog( "Can't open User Profile file \'$UsPrFile\'. ($!)." );
			print STDERR "Can't open User Profile file \'$UsPrFile\'. ($!).\n";
			$UserProfile{$u_id}{0} = 0;
		    }

}

sub areafix()
{

    my ( $chat, $listtext, $invite, $desc, $ch_id );

	unless( defined( $areaname{$chatid} ) ) {
# help message
	    if ( $msgtext =~ /[\r\n\s]*\%help[\r\n\s]*/i ){
		writelog( "Help request from $fromname ($chatid) found." );
		undef $error_message;
		$result = eval { $api->sendMessage (
				{
				    chat_id => $chatid,
				    text    => $helptext
				} )
				} or $error_message = $api->parse_error->{msg};
		writelog("sendMessage got error: $error_message") if defined($error_message);
	    }
# help message end

# list message 
	    $listtext = '';
	    $invite = '';
#	    undef $error_message;
	    if ( $msgtext =~ /[\r\n\s]*\%list[\r\n\s]*/i ){
		writelog( "List request from $fromname ($chatid) found." );
		undef $error_message;
		$result = eval { $api->sendMessage (
				{
				    chat_id => $chatid,
				    text    => 'The list of echo conferences is being prepared. Wait please. It can take some time.'
				} )
				} or $error_message = $api->parse_error->{msg};
		writelog("sendMessage got error: $error_message") if defined($error_message);

		foreach $ch_id ( sort {$areaname{$a} cmp $areaname{$b}} keys %areaname ) {
		    undef $error_message;
		    $result = eval { $chat = ( $api->getChat ({
			    chat_id => $ch_id
			}))
		    } or $error_message = $api->parse_error->{msg};
		    undef $invite;
		    undef $desc;
		    $invite = ( $chat->{result}->{invite_link} );
		    $desc = ( $chat->{result}->{description} );
		    print STDERR "$error_message\n" if defined $error_message && $error_message ne '';
		    writelog ( "$areaname{$ch_id} \($ch_id\)" );
		    writelog( $error_message ) if defined $error_message && $error_message ne '';
		    $listtext .= "$areaname{$ch_id}";
		    $listtext .= " - $desc" if defined $desc;
		    $listtext .= "\n$invite" if defined $invite;
		    $listtext .= "\n\n";
		    if( length($listtext) > 3800 ){
			$listtext .= "\*\*\* To be continued. \*\*\*\n";
			undef $error_message;
			$result = eval { $api->sendMessage (
				{
				    chat_id => $chatid,
				    text    => $listtext
				} )
				} or $error_message = $api->parse_error->{msg};
#			sleep 2;
			writelog("sendMessage got error: $error_message") if defined($error_message);
			$listtext = '';
		    }
#		    sleep 1;
		}
		undef $error_message;
		$result = eval { $api->sendMessage (
				{
				    chat_id => $chatid,
				    text    => $listtext
				} )
				} or $error_message = $api->parse_error->{msg};
		sleep 1;
		writelog("sendMessage got error: $error_message") if defined($error_message);
	    }
# list message end

# query
	    $listtext = "List of areas which you are currently linked to:\n";
	    $invite = '';
	    if ( $msgtext =~ /[\r\n\s]*(\%query|\%LINKED)[\r\n\s]*/i ){
		writelog( "Query request from $fromname ($chatid) found." );
		undef $error_message;
		$result = eval { $api->sendMessage (
				{
				    chat_id => $chatid,
				    text    => 'The list of echo conferences is being prepared. Wait please. It can take some time.'
				} )
				} or $error_message = $api->parse_error->{msg};
		foreach $ch_id ( sort {$areaname{$a} cmp $areaname{$b}} keys %areaname ) {
		    undef $error_message;
		    $result = eval { $invite = ( $api->getChatMember ({
			    chat_id => $ch_id,
			    user_id => $chatid
			})->{result}->{status} )
		    } or $error_message = $api->parse_error->{msg};
		    print STDERR "$error_message\n" if defined $error_message && $error_message ne '';
		    writelog ( $areaname{$ch_id} );
		    writelog( $error_message ) if defined $error_message && $error_message ne '';
		    $listtext .= "$areaname{$ch_id} - $invite\n" if $invite =~ /^creator|administrator|member$/i;
		    sleep 1;
		}
		undef $error_message;
		$result = eval { $api->sendMessage (
				{
				    chat_id => $chatid,
				    text    => $listtext
				} )
				} or $error_message = $api->parse_error->{msg};
		sleep 1;
		writelog("sendMessage got error: $error_message") if defined($error_message);
	    }
# qery end

# unlinked
	    $listtext = "List of unlinked areas:\n";
	    $invite = '';
	    if ( $msgtext =~ /[\r\n\s]*\%unlinked[\r\n\s]*/i ){
		undef $error_message;
		$result = eval { $api->sendMessage (
				{
				    chat_id => $chatid,
				    text    => 'The list of echo conferences is being prepared. Wait please. It can take some time.'
				} )
				} or $error_message = $api->parse_error->{msg};
		writelog( "Unlinked request from $fromname ($chatid) found." );
		foreach my $ch_id ( sort {$areaname{$a} cmp $areaname{$b}} keys %areaname ) {
		    undef $error_message;
		    $result = eval { $invite = ( $api->getChatMember ({
			    chat_id => $ch_id,
			    user_id => $chatid
			})->{result}->{status} )
		    } or $error_message = $api->parse_error->{msg};
		    print STDERR "$error_message\n" if defined $error_message && $error_message ne '';
		    writelog ( $areaname{$ch_id} );
		    writelog( $error_message ) if defined $error_message && $error_message ne '';
		    $listtext .= "$areaname{$ch_id} - $invite\n" if $invite =~ /^restricted|left|kicked$/i;
		    sleep 1;
		}
		undef $error_message;
		$result = eval { $api->sendMessage (
				{
				    chat_id => $chatid,
				    text    => $listtext
				} )
				} or $error_message = $api->parse_error->{msg};
		sleep 1;
		writelog("sendMessage got error: $error_message") if defined($error_message);
	    }

# unlinked end

# $UsersProfileDir
#	    my $UsPrFile;
	    if ( $msgtext =~ /[\r\n\s]*\%Userprofile[\r\n\s]*/i ) {
		unless( defined( $UsersProfileDir ) ) {
		    $listtext = 'User Profiles Disabled.';
		} else {
		    readuserprofile( $chatid ) unless defined $UserProfile{$chatid}{0};
		    if( $msgtext =~ /\%RealName\s+(\S[^\r\n]+)/i ) {
			$UserProfile{$chatid}{realname} = $1;
			$UserProfile{$chatid}{0} = 2;
			if( $UserProfile{$chatid}{realname} =~ /\$del/i ) {
			    undef $UserProfile{$chatid}{realname};
			}
		    }
		    if ( $msgtext =~ /\%TagLine\s+(\S[^\r\n]+)/i ) {
			$UserProfile{$chatid}{tagline} = $1;
			$UserProfile{$chatid}{tagline} = substr( $UserProfile{$chatid}{tagline}, 0, 74 ) if length( $UserProfile{$chatid}{tagline} ) > 74;
			$UserProfile{$chatid}{0} = 2;
			if( $UserProfile{$chatid}{tagline} =~ /\$del/i ) {
			    undef $UserProfile{$chatid}{tagline};
			}
		    }
		    if ( $msgtext =~ /\%Tearline\s+(\S[^\r\n]+)/i ) {
			$UserProfile{$chatid}{tearline} = $1;
			$UserProfile{$chatid}{tearline} = substr( $UserProfile{$chatid}{tearline}, 0, 74 ) if length( $UserProfile{$chatid}{tearline} ) > 74;
			$UserProfile{$chatid}{0} = 2;
			if( $UserProfile{$chatid}{tearline} =~ /\$del/i ) {
			    undef $UserProfile{$chatid}{tearline};
			}
		    }
		    if ( $msgtext =~ /\%FidoAddr\s+(\S[^\r\n]+)/i ) {
			$UserProfile{$chatid}{fidoaddr} = $1;
			$UserProfile{$chatid}{0} = 2;
			if( $UserProfile{$chatid}{fidoaddr} =~ /\$del/i ) {
			    undef $UserProfile{$chatid}{fidoaddr};
			}
		    }
		    if ( $msgtext =~ /\%Greeting\s+(\S[^\r\n]+)/i ) {
#			print "\%Greeting\n";
			$UserProfile{$chatid}{greeting} = $1;
			$UserProfile{$chatid}{0} = 2;
			if( $UserProfile{$chatid}{greeting} =~ /\$del/i ) {
			    undef $UserProfile{$chatid}{greeting};
			}
		    }
		    if ( $msgtext =~ /\%Conclusion\s+(\S[^\r\n]+)/i ) {
			$UserProfile{$chatid}{conclusion} = $1;
			$UserProfile{$chatid}{0} = 2;
			if( $UserProfile{$chatid}{conclusion} =~ /\$del/i ) {
			    undef $UserProfile{$chatid}{conclusion};
			}
		    }
		}
		$listtext = "Your \%UserProfile\n\%RealName ";
		$listtext .= $UserProfile{$chatid}{realname} if defined $UserProfile{$chatid}{realname};
		$listtext .= "\n\%FidoAddr ";
		$listtext .= $UserProfile{$chatid}{fidoaddr} if defined $UserProfile{$chatid}{fidoaddr};
		$listtext .= "\n\%Greeting ";
		$listtext .= $UserProfile{$chatid}{greeting} if defined $UserProfile{$chatid}{greeting};
		$listtext .= "\n\%Conclusion ";
		$listtext .= $UserProfile{$chatid}{conclusion} if defined $UserProfile{$chatid}{conclusion};
		$listtext .= "\n\%TagLine ";
		$listtext .= $UserProfile{$chatid}{tagline} if defined $UserProfile{$chatid}{tagline};
		$listtext .= "\n\%Tearline ";
		$listtext .= $UserProfile{$chatid}{tearline} if defined $UserProfile{$chatid}{tearline};
		$listtext .= "\n\n";

		my $UsPrFile = File::Spec->catfile( $UsersProfileDir, "$chatid.conf" );
		if ( $UserProfile{$chatid}{0} == 2 ) {
		    if ( open( UPFH, ">$UsPrFile" ) ) {
		        print UPFH $listtext;
		        close( UPFH );
		    } else {
		        writelog( "Can't open User Profile file \'$UsPrFile\'. ($!)." );
		        print STDERR "Can't open User Profile file \'$UsPrFile\'. ($!).\n";
		    }
		}
		unless ( defined $UserProfile{$chatid}{realname} ) {
		 undef $error_message;
		 $result = eval { $api->sendMessage (
				{
				    chat_id => $chatid,
				    text    => 'ERROR! Realname MUST be defined!',
				    entities => [{ offset => 0,
						length => 6,
						type => 'bold' },
						{ offset => 16,
						length => 4,
						type => 'bold' }
						]
				} )
				} or $error_message = $api->parse_error->{msg};
		 writelog("sendMessage got error: $error_message") if defined($error_message);
		 sleep 1;
		}
		undef $error_message;
		$result = eval { $api->sendMessage (
				{
				    chat_id => $chatid,
				    text    => decode( 'cp866', $listtext )
				} )
				} or $error_message = $api->parse_error->{msg};
		writelog("sendMessage got error: $error_message") if defined($error_message);
		sleep 2;
	    }
# $UsersProfileDir end
	}
}

sub doquote()
{
    return unless defined $replytext;

    my($onestr);
    my @strs = split "\r", $replytext;

    $replytext = '';
    foreach $onestr ( @strs ) {
	if ( $onestr =~ /^(\s?[^ \>]{1,3}\>+)\s(.*$)/ ) {
#	    $onestr = "$1\> $2";
	    $onestr = '';
	} else {
	    $onestr = " $quotechars\> $onestr\r" if $onestr ne '';
	}
	$replytext .= "$onestr\r";
    }
    $replytext =~ s/\r\r/\r/g while $replytext =~ /\r\r/;
#    $replytext =~ s/(\r\n?)[ ]?/$1 $quotechars\> /g if defined $replytext;
#    $replytext =~ s/[\-\.\=\,\+\/]{2}CC\:\s*([^\r\n\s]+?)\s*[\-\.\=\,\+\/]{2}//i if defined $replytext;
}

sub read_from_groups()
{
    my $ccarea;
    undef $chatid;
    undef $chattitle;
    undef $msgtext;
    undef $subj;
    undef $replytext;
    undef $forward_from_chat_title;
    undef $fwd_fromname;
    undef $tg_msg_id;
    undef $reply_msgid;
    
	$chatid = ( $api->getUpdates->{result}[$u]{message}{chat}->{id} );
	$chattitle = ( $api->getUpdates->{result}[$u]{message}{chat}->{title} );
	if ( defined($chattitle) ) {
	    if ( !defined( $areaname{$chatid} ) ) {
		$chattitle = encode('cp866',$chattitle);
		writelog( "Unknown group \'$chattitle\' ($chatid)" );
		print "Unknown group \'$chattitle\' ($chatid)\n";
		undef $error_message;
		$result = eval { $api->sendMessage (
				{
				    chat_id => $chatid,
				    text    => "FORBIDDEN!\nThis bot can not be used in this group.",
				    entities => [{ offset => 0,
						    length => 10,
						    type => 'bold' }
						]
				} )
				} or $error_message = $api->parse_error->{msg};
		writelog("sendMessage got error: $error_message") if defined($error_message);
#		undef $chattitle;
		return;
	    }
	}
	$tg_msg_id = ( $api->getUpdates->{result}[$u]{message}->{message_id} );
	my $rm_id = ( $api->getUpdates->{result}[$u]{message}{reply_to_message}->{message_id} );
	$reply_msgid = sprintf ( "$address %08x", $rm_id ) if defined $rm_id;
#	sprintf ("\001REPLY: $msg_fromaddr %08x\r", $reply_msgid )
#	print "$chattitle\n";
#	$forward_from_chat_title = ( $api->getUpdates->{result}[$u]{message}{forward_from_chat}->{title} );

	get_from_names();
	readuserprofile( $from_user_id ) unless defined $UserProfile{$from_user_id}{0};

#	gone_or_new_users() if defined $chattitle;
	if( defined( ($api->getUpdates->{result}[$u]{message}->{new_chat_members}) ) ) {
	    print "New user message shouldn't be sent.\n";
	    writelog( "\'$chattitle\' New user message shouldn't be sent." );
	    if ( $UserProfile{$from_user_id}{0} == 0 ){
		undef $error_message;
		$result = eval { $api->sendMessage (
		        {
			    chat_id => $chatid,
			    text    => "WARNING!\nTo write to Fido echo conferences, you need to register with \@Fido2Telebot by sending the command \%USERPROFILE in the chat with \@Fido2Telebot. Use \%HELP in the chat with \@Fido2Telebot for more details.",
			    reply_to_message_id => $tg_msg_id,
			    entities => [{ offset => 0,
					    length => 8,
					    type => 'bold' }
					]
			} )
			} or $error_message = $api->parse_error->{msg};
		writelog("sendMessage got error: $error_message") if defined($error_message);
	    }
	    return;
	}
	if( defined( ( $api->getUpdates->{result}[$u]{message}->{left_chat_member}->{first_name} ) ) ) {
	    print "Gone user message shouldn't be sent.\n";
	    writelog( "\'$chattitle\'Gone user message shouldn't be sent." );
	    return;
	}

	$msgtext = ( $api->getUpdates->{result}[$u]{message}->{text} ) if !defined $msgtext;
	
#	if( defined( $msgtext ) ){
#	    $msgtext =~ s/\N{U+F09F}\N{U+9889}/\;\-\)/g;
#	    $msgtext =~ s/\xF0\x9F\x98\x8A/\:\-\)/g;
#	    $msgtext =~ s/\xf0\x9f\x98\x9e/\:\-\(/g;
#	    print "Error open file \'$temp_dir\/utf-8\/".sprintf( "%08x",time() ).".\'($!)" unless open(FHUTF, ">$temp_dir\/utf\-8\/".sprintf( "%08x",time() ) );
#	    binmode(FHUTF,':utf8');
#	    print(FHUTF "$msgtext\r\n--------------\r\n");
#	    close(FHUTF);
#	}
	
	$replytext = ( $api->getUpdates->{result}[$u]{message}{reply_to_message}->{text} );
#	$reply_msgid = ( $api->getUpdates->{result}[$u]{message}{reply_to_message}->{message_id} );
	$forward_from_chat_title = encode( 'cp866', ( $api->getUpdates->{result}[$u]{message}{forward_from_chat}->{title} ) );

	$msgtext = encode( 'cp866', $msgtext ) if defined $msgtext;
	$replytext = encode( 'cp866', $replytext ) if defined $replytext;
	$caption = ( $api->getUpdates->{result}[$u]{message}->{caption} );
	$msgtext = encode('cp866',$caption) if defined $caption;
	get_entities();
	$msgtext =~ s/\r?\n/\r/g if defined $msgtext;
	$replytext =~ s/\r?\n/\r/g if defined $replytext;
	areafix();
	set_subj();
	get_to_name();
	
#	$replytext = " $quotechars\> $replytext\r" if defined $replytext;
	
	get_file();
	
	$msgtext = "Forwarded from chat: $forward_from_chat_title\r$msgtext" if defined $forward_from_chat_title;
	$msgtext = "Forward from: $fwd_fromname\r$msgtext" if defined $fwd_fromname;
	
#	$replytext = " $quotechars\> $replytext" if defined $replytext;
#	$replytext =~ s/(\r\n?)[ ]?/$1 $quotechars\> /g if defined $replytext;
	doquote();
	$replytext =~ s/[\-\.\=\,\+\/]{2}CC\:\s*([^\r\n\s]+?)\s*[\-\.\=\,\+\/]{2}//i if defined $replytext;
	
	$msgtext = "$replytext\r$msgtext" if defined $replytext;
	$msgtext .= "\r";
#--- User Profile -->
	my ( $rn, $greet, $tnf, $conc, $sfn );
	if ( $UserProfile{$from_user_id}{0} != 0 ){
	    $toname =~ /^([^ \.]+).*$/;
	    $tnf = $1;
	    $fromname =~ /^([^ \.]+).*$/;
	    $sfn = $1;
	    if ( defined( $UserProfile{$from_user_id}{greeting} ) ) {
		$greet = $UserProfile{$from_user_id}{greeting};
		$greet =~ s/\$tnf/$tnf/ig; # 
		$greet =~ s/\$sfn/$sfn/ig; # 
		$greet =~ s/\$rfn/$toname/ig;#
		$greet =~ s/\$fn/$fromname/ig; #
		$greet =~ s/\$cr/\r/ig; #
		$msgtext = "$greet\r\r$msgtext";
	    }
	    if ( defined $UserProfile{$from_user_id}{realname} ) {
		$rn = "\001RealName: $UserProfile{$from_user_id}{realname}";
		$rn .= " $UserProfile{$from_user_id}{fidoaddr}" if defined $UserProfile{$from_user_id}{fidoaddr};
		writelog($rn);
		$msgtext = "$rn\r$msgtext";
	    }
	    if ( defined( $UserProfile{$from_user_id}{conclusion} ) ){
		$conc = $UserProfile{$from_user_id}{conclusion};
		$conc =~ s/\$fn/$fromname/ig; #
		$conc =~ s/\$rfn/$toname/ig; #
		$conc =~ s/\$tnf/$tnf/ig; #
		$conc =~ s/\$sfn/$sfn/ig; #
		$conc =~ s/\$cr/\r/ig; #
		$msgtext .= "\r$conc\r\r";
	    }
	    $msgtext .= "... $UserProfile{$from_user_id}{tagline}\r" if defined $UserProfile{$from_user_id}{tagline};
	    my $tear_l;
	    if ( defined( $UserProfile{$from_user_id}{tearline} ) ) {
		$tear_l = "--- $UserProfile{$from_user_id}{tearline}\r";
	    } else {
		$tear_l = $tearline;
	    }
	    $msgtext .= $tear_l;
	    
	} else {
	    writelog("WARNING: Unknown user: $fromname ($from_user_id).");
	    print STDERR "WARNING: Unknown user: $fromname ($from_user_id).\n";
	    if ( defined( $areaname{$chatid} ) ) {
		my $an = $areaname{$chatid};
		if ( $options{$an} !~ /AllowUnreg/i ){
#		    undef $error_message;
#		    $result = eval { $api->sendMessage (
#			    {
#				chat_id => $chatid,
#				text    => "WARNING!\nTo write to Fido echo conferences, you need to register on BBS. Use the command %USERPROFILE in the chat with \@Fido2Telebot. Use %HELP in the chat with \@Fido2Telebot for more details.",
#				reply_to_message_id => $tg_msg_id,
#				entities => [{ offset => 0,
#						length => 8,
#						type => 'bold' }
#					    ]
#			    } )
#			    } or $error_message = $api->parse_error->{msg};
#		    writelog("sendMessage got error: $error_message") if defined($error_message);
		    if ( $options{$an} =~ /RegisteredOnly/i ){
			writelog( "$an is Registered Only. Unknown user $fromname ($from_user_id) is not permited to write messages to Fido." );
			return;
		    }
		}
	    }
#
	}
#<--- End of User Profile --

	unless( defined( $areaname{$chatid} ) ) {
	    if ( $toname =~ /^Ping$/i ) {
		$fromname = $from_user_id;
	    } else {
		$msgtext = "\001REPLYTO $address $from_user_id\r\n$msgtext";
		$msgtext = "\x01RFC-Reply-To: $chatid\@f$af\.n$an\.z$az.fidonet.org\r$msgtext"
	    }
	} else {
		$msgtext = "\001REPLYTO $address $from_user_id\r\n$msgtext";
	}
# <---REPLYTO
	$msgtext = "\001CHRS: CP866 2\r\001TGUID: $from_user_id\r$msgtext";

# carbon copy, cross post
	my ( %XC_areaname, $xc_msgtext );
	my $i = 0;
	while( $msgtext =~ /([\-\.\=\,\+\/\$])\1[CX]C\:\s*([^\r\n\s]+?)\s*\1\1\r?/i ){
	    $XC_areaname{$i} = uc($2);
	    $msgtext =~ s/([\-\.\=\,\+\/\$])\1[CX]C\:\s*([^\r\n\s]+?)\s*\1\1\r?//i;
	    $i++;
	}
	$xc_msgtext = $msgtext;
	$xc_msgtext =~ s/\x01[^r]+\r//g;
	$xc_msgtext =~ s/\r\.\.\. [^\r]+//g;
	$xc_msgtext =~ s/\r--- [^\r]+//g;
#	$xc_msgtext =~ s/\r\r/\r/g while $xc_msgtext =~ /\r\r/;
	$xc_msgtext =~ s/\r/\n/g;
	while( $i > 0) {
	    $i--;
	    $ccarea = $XC_areaname{$i};
	    if( defined( $tags_id{$ccarea} ) ) {
		writelog("Area \'$ccarea\' is legal. Sending a copy.");
		sendaspkt( $fromname, $toname, $address, $toaddr, $msgtext,
			     $subj, $ccarea ) if defined( $msgtext );
		undef $error_message;
		$result = eval { $api->sendMessage (
				{
				chat_id => $tags_id{$ccarea},
				text    => "--From: $fromname $address--\n".
					    "--To: $toname\n".
					    "--Subj: $subj--\n".
					    sprintf("$address %08x\n\n",
						$tg_msg_id).
					    decode( 'cp866',$xc_msgtext ),
				disable_notification => 1
				} )
				} or $error_message = $api->parse_error->{msg};
		writelog("copyMessage got error: $error_message") if defined($error_message);
		sleep 3;
	    } else{
		writelog("Warning: An attempt to send a copy to an unknown area.");
		undef $error_message;
		$result = eval { $api->sendMessage (
				{
				    chat_id => $from_user_id,
				    text    => "FORBIDDEN!\nAn attempt to send a copy to an unknown area \'$ccarea\'.",
				    entities => [{ offset => 0,
						    length => 10,
						    type => 'bold' }
						]
				} )
				} or $error_message = $api->parse_error->{msg};
		writelog("sendMessage got error: $error_message") if defined($error_message);
	    }
	    
	} # end foreach
	
#	    $ccarea = uc($1);
#	    $msgtext =~ s/[\-\.\=\,\+\/\$]{2}CC\:\s*([^\r\n\s]+?)\s*[\-\.\=\,\+\/\$]{2}//i;
#	    writelog("Carbon Copy to area \'$ccarea\' found");
#	    if( defined( $tags_id{$ccarea} ) ) {
#		writelog("Area \'$ccarea\' is legal. Sending a copy.");
#		sendaspkt( $fromname, $toname, $address, $toaddr, $msgtext, $subj, $ccarea ) if defined( $msgtext );
#		undef $error_message;
#		$result = eval { $api->copyMessage (
#				{
#				    chat_id => $tags_id{$ccarea},
#				    from_chat_id => $chatid,
#				    message_id => $tg_msg_id,
#				    disable_notification => 1
#				    
#				} )
#				} or $error_message = $api->parse_error->{msg};
#		writelog("copyMessage got error: $error_message") if defined($error_message);
#		sleep 3;
#	    } else{
#		$result = eval { $api->sendMessage (
#				{
#				    chat_id => $from_user_id,
#				    text    => "FORBIDDEN!\nAn attempt to send a copy to an unknown area \'$ccarea\'.",
#				    entities => [{ offset => 0,
#						    length => 10,
#						    type => 'bold' }
#						]
#				} )
#				} or $error_message = $api->parse_error->{msg};
#		writelog("sendMessage got error: $error_message") if defined($error_message);
#		writelog("Warning: An attempt to send a copy to an unknown area.");
#	    }
#	}
# carbon copy

	sendaspkt( $fromname, $toname, $address, $toaddr, $msgtext, $subj, $areaname{$chatid} ) if defined( $msgtext );
}

sub read_from_telegram()
{
	undef $error_message;

	print "Getting updates from Telegram...\n";
#	writelog("Getting updates from Telegram...");

	$result = eval { $api->getUpdates(
#	    allowed_updates => ['message', 'edited_message'],
#	    timeout => 30
	) } or $error_message = $api->parse_error->{msg};
	if ( defined( $error_message ) ){
		writelog( "ERROR: getUpdates got \'$error_message\'" );
		print STDERR "ERROR: getUpdates got \'$error_message\'\n";
		rm_lock();
		exit;
	}

	undef $error_message;
	$result = eval {
	    $updcount = @{( $api->getUpdates->{result} )}
	} or $error_message = $api->parse_error->{msg};
	if ( defined( $error_message ) ){
	    if( length($error_message) > 1 ){
	    print( STDERR "ERROR! Update count got: $error_message\n");
	    writelog("Update count got: $error_message");
	    return;
	    }
	}

	print("Updates count: $updcount\n");
	writelog("Updates count: $updcount") if $updcount > 0;

	for ($u = 0; $u < $updcount; $u++ ) {
		print "update number: $u\n";
		writelog( "update number: $u" );
		$updateid = ( $api->getUpdates->{result}[$u]->{update_id} );
		next if !defined ( $api->getUpdates->{result}[$u]->{message} );
		read_from_groups();
		
#		sleep 1;
	} # --- End for($u = 0; $u < $updcount; $u++ ) ---

	if ( $updcount > 0 ) {
#		$api->getUpdates( { offset => $updateid + 1 } );
		$result = eval { $api->getUpdates( { offset => $updateid + 1 } ) } or $error_message = $api->parse_error->{msg};
		if ( defined( $error_message ) ){
			writelog( "ERROR: getUpdates \(\{ offset => $updateid \+ 1 \}\) got \'$error_message\'" );
			print STDERR "ERROR: getUpdates \(\{ offset => $updateid \+ 1 \}\) got \'$error_message\'\n";
#			rm_lock();
#			exit;
		}
	}

	print "Getting updates from Telegram Done.\n";
#	writelog("Getting updates from Telegram Done.");
}

#sub createlock()
#{
#	if(open($LFH,">$lockfile")) {
#	    print $LFH $$;
#	    writelog("$lockfile created. PID: $$");
#	    print "Lock file created. PID: $$\n";
#	    close( $LFH );
#	} else {
#	    print STDERR "Can't create $lockfile. ($!).\n";
#	    writelog("Can't create $lockfile. ($!).");
#	    exit;
#	}
#}

sub rm_lock()
{
#	close($LFH);
	if( !unlink $lockfile ){
	    print STDERR "Can't delete $lockfile. ($!)\n";
	    writelog("Can't delete $lockfile. ($!)");
	}else{
	    print "[$$] Lock file deleted.\n";
	    writelog("$lockfile deleted.");
	}
	if( -e $exitflag ) {
#	    writelog("Exit flag found.");
#	    print "[$$] Exit flag found.\n";
	    if( !unlink $exitflag ){
		print STDERR "Can't delete $exitflag. ($!)\n";
		writelog("Can't delete $exitflag. ($!)");
	    }else{
		print "[$$] Exit flag deleted.\n";
		writelog("$exitflag deleted.");
	    }
	}

}

my $pid;

sub check_lock()
{
    $pid = File::Flock::Tiny->write_pid($lockfile) or do {
	print("$lockfile found and can not be deleted ($!). Exitting.\n");
	writelog("$lockfile found. Exitting.");
	exit;
    };
    print " [***] PID: $$ $0 [***]\n";
}

sub configure()
{
    read_conf();
    check_errors();
    readchannelsconf();
}

# <---- Start ---->

configure();
#readuserconf();

if ( !defined( $toss ) && !defined( $scan ) ) {
	print STDERR "\n--toss or --scan shold be defined.\n\n$usage";
	writelog("--toss or --scan should be defined. Nothing to do. Exitting.");
	exit;
}

check_lock();
#$ENV{'MOJO_INACTIVITY_TIMEOUT'} = '60';

$api = WWW::Telegram::BotAPI->new(
    token => $bot_token,
    api_url => $api_url
);

#$api->agent->inactivity_timeout (45);

#$cycle_time = 6;
#$cycle_count = 10;
our $c = 1;

sub scanit()
{
    while ($c == 1) {
	read_from_telegram();
	$c = 0 && writelog('Exit flag found. Exitting Scan.') if -e $exitflag;
	$c = 0 && writelog("Config file was changed. Exitting Scan.") if conf_stat();
    }
}

sub tossit()
{
    while ($c == 1) {
        readpktdir();
        print "[$$] Toss Waiting $cycle_time sec. for next cycle.\n" if $c == 1;
	sleep $cycle_time if $c == 1;
	$c = 0 && writelog('Exit flag found. Exitting Toss.') if -e $exitflag;
	$c = 0 && writelog('Config file was changed. Exitting Toss.') if conf_stat();
    }

}


    my ($ths, $tht, $thr_num);
    my $threads = 0;
    my $cm = 1;

while( $cm ) {
    $ths = threads->create(\&scanit) if $scan;
    $tht = threads->create(\&tossit) if $toss;
    $thr_num = threads->list(threads::running);
    print "threads = $thr_num\n";
    writelog("Started with $thr_num threads.");
    $ths->join() if $scan; 
    $tht->join() if $toss;
    writelog('Config changed...') && configure() if conf_stat();
    $pid = File::Flock::Tiny->write_pid($lockfile) or do {
	print("$lockfile found and is actual...\n");
	writelog("$lockfile found and is actual...");
	};
    $thr_num = threads->list(threads::running);
    print "thrreads = $thr_num\n";
    $cm = 0 if -e $exitflag;
}
 
#for( my $c = $cycle_count; $c > 0; $c-- ){
#    writelog("Cycle count: $c.");
#    readpktdir() if $toss;
#    read_from_telegram() if $scan;
#    print "Waiting $cycle_time sec. for next cycle.\n" if $c > 1;
#    sleep $cycle_time if $c > 1;
#}

# <---- The End ---->
rm_lock();
writelog("Normal exit.");
# <---- The End ---->
